/***************************************************************************
 *   Copyright (C) 2008 by Thomas Schildknecht <danakil@gmail.com>         *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA .        *
 *                                                                         *
 ***************************************************************************/

#include "activewindowcontrol.h"

// KDE
#include <KIcon>
#include <kworkspace/kworkspace.h>
#include <KDebug>

#include <KWindowSystem> 
#include <NETRootInfo>
#include <QX11Info>  
#include <KActionSelector> 
#include <QListWidgetItem>


#define ACTION_CLOSE            1000
#define ACTION_MAXIMIZE         1001
#define ACTION_MINIMIZE         1002
#define ACTION_SHADE            1003
#define ACTION_KEEPABOVE        1004
#define ACTION_KEEPBELOW        1005
#define SEPARATOR               1006
#define FIRST_ACTION            1000
#define LAST_ACTION             1005
#define SEPARATOR_WIDTH         8
#define PLANAR_WIDTH            38
#define PLANAR_HEIGHT           64

ActiveWindowControl::ActiveWindowControl(QObject *parent, const QVariantList &args) :     
    Plasma::Applet(parent, args),        
    m_configurationDialog(0),
    separatorsAdded(0),
    isOnPanel(true),
    isVerticalPref(false),
    m_activeWindow(0)
{
    setHasConfigurationInterface(true);
    
    // Qt::WindowFlags flags = this->windowFlags ();
    // flags |= Qt::X11BypassWindowManagerHint;
    // this->setWindowFlags(flags);
    // this->setFocusPolicy(Qt::NoFocus);
}

void ActiveWindowControl::init()
{
    KConfigGroup cg = config();
    actionIdList = cg.readEntry("idActionsSelected", actionIdList);
    isVerticalPref = cg.readEntry("verticalLayout", isVerticalPref);

    // Set some default buttons if the user prefs are empty
    if(actionIdList.isEmpty()) {
        actionIdList<<"1002"<<"1001"<<"1006"<<"1000";
        cg.writeEntry("idActionsSelected", actionIdList); 
        emit configNeedsSaving();
    }
    
    if(formFactor() == Plasma::Planar || formFactor() == Plasma::MediaCenter)
        isOnPanel=false;
    else isOnPanel=true;
  
    //connect (KWindowSystem::self(), SIGNAL(activeWindowChanged (WId)), this, SLOT(slotNewActiveWindow(WId)));

    // Set the layout
    m_layout = new QGraphicsLinearLayout(0);
    m_layout->setContentsMargins(0, 0, 0, 0);
    m_layout->setSpacing(0);
     
    if(isVerticalPref) m_layout->setOrientation(Qt::Vertical);  
    else m_layout->setOrientation(Qt::Horizontal);
        
    setLayout(m_layout);

    addAllItems();
    updateConstraints();
}

ActiveWindowControl::~ActiveWindowControl()
{
    if(m_configurationDialog) delete m_configurationDialog;
    removeAllActionButtons();
}

void ActiveWindowControl::constraintsEvent(Plasma::Constraints constraints)
{  
    if (constraints & Plasma::SizeConstraint || constraints & Plasma::FormFactorConstraint) 
    {      
        if(formFactor() == Plasma::Planar || formFactor() == Plasma::MediaCenter) {                                 
            if(!isVerticalPref)
            setMinimumSize((iconList.size()*PLANAR_WIDTH)+(separatorsAdded*SEPARATOR_WIDTH),PLANAR_HEIGHT);
            else
            setMinimumSize(PLANAR_HEIGHT, (iconList.size()*PLANAR_WIDTH)+(separatorsAdded*SEPARATOR_WIDTH));
   
            isOnPanel=false;
        }
        else if(formFactor() == Plasma::Horizontal) {
            m_layout->setOrientation(Qt::Horizontal);          
            setBackgroundHints(NoBackground);  
            setMinimumSize(0, 0);
            int h= containment()->size().height();
            setMaximumSize(h*iconList.size() + (separatorsAdded* SEPARATOR_WIDTH) -
                (iconList.size()*6), h); // [...*6]:reduce the gap between items
             
            isOnPanel=true;
        }
        else if(formFactor() == Plasma::Vertical) {          
            m_layout->setOrientation(Qt::Vertical);
            setBackgroundHints(NoBackground);  
            setMinimumSize(0, 0);
            int w= containment()->size().width();
            setMaximumSize(w, w*iconList.size() + (separatorsAdded*SEPARATOR_WIDTH) -
                (iconList.size()*6) ); // [...*6]:reduce the gap between items
            isOnPanel=true;
        }
        m_layout->updateGeometry(); // ???
    }
}

void ActiveWindowControl::showConfigurationInterface()
{
    if (m_configurationDialog == 0) 
    {
        m_configurationDialog = new KDialog;
        m_configurationDialog->setCaption(i18n("ActiveWindowControl Configuration"));

        QWidget *widget = new QWidget;
        uiConfig.setupUi(widget);
        m_configurationDialog->setMainWidget(widget);
        m_configurationDialog->setButtons(KDialog::Ok | KDialog::Cancel | KDialog::Apply);

        connect (uiConfig.winactionselector, SIGNAL(added(QListWidgetItem *)), this, SLOT(slotActionAdded(QListWidgetItem *)));
        connect (uiConfig.winactionselector, SIGNAL(removed(QListWidgetItem *)), this, SLOT(slotActionRemoved(QListWidgetItem *)));
        connect (m_configurationDialog, SIGNAL(applyClicked()), this, SLOT(configAccepted()));
        connect (m_configurationDialog, SIGNAL(okClicked()), this, SLOT(configAccepted()));

        // Hide the "use a vertical layout" checkbox if the plasmoid is in a panel
        if(isOnPanel) uiConfig.checkBoxVerticalLayout->hide();
        else {
            if(isVerticalPref) uiConfig.checkBoxVerticalLayout->setCheckState(Qt::Checked);
            else uiConfig.checkBoxVerticalLayout->setCheckState(Qt::Unchecked);
        }

        // Initialize the KActionSelector
        //
        int idAction;
        QStringList textData;
        textData<<i18n("Close") <<"awc-close"       
                <<"Maximize"    <<"awc-maximize"    
                <<"Minimize"    <<"awc-minimize"    
                <<"Shade"       <<"awc-shade"        
                <<"Keep above"  <<"awc-keepabove"   
                <<"Keep below"  <<"awc-keepbelow"; 
                                                
        for(int i=0; i<actionIdList.size(); i++) {         // add the user-selected actions        
            idAction=actionIdList[i].toInt();
            
            if(idAction>=FIRST_ACTION && idAction<=LAST_ACTION)
                new QListWidgetItem(KIcon(textData[((idAction-FIRST_ACTION)*2)+1]), 
                                    textData[(idAction-FIRST_ACTION)*2], 
                                    uiConfig.winactionselector->selectedListWidget(),idAction); 
            else if(idAction==SEPARATOR) 
                new QListWidgetItem("--- space ---", uiConfig.winactionselector->selectedListWidget(),SEPARATOR);                                     
        }                                     
        
        for(int i=FIRST_ACTION;i<=LAST_ACTION;i++) {     // add the other actions to the available list
        
              if(! actionIdList.contains(QString().setNum(i)))
                  new QListWidgetItem(KIcon(textData[((i-FIRST_ACTION)*2)+1]), 
                      textData[(i-FIRST_ACTION)*2], 
                      uiConfig.winactionselector->availableListWidget(),i);                 
        }
        new QListWidgetItem("--- space ---", uiConfig.winactionselector->availableListWidget(),SEPARATOR);

    }
    m_configurationDialog->show();  
}

void ActiveWindowControl::slotActionAdded(QListWidgetItem * item)
{
    if(item->type()==SEPARATOR)
        new QListWidgetItem(QIcon(""),"--- space ---", 
                            uiConfig.winactionselector->availableListWidget(),SEPARATOR); 
}

void ActiveWindowControl::slotActionRemoved(QListWidgetItem * item)
{
    if(item->type()==SEPARATOR)
        delete item;
}

void ActiveWindowControl::configAccepted()
{
    KConfigGroup cg = config();
    
    // Save the KActionSelector selected list
    QString value;
    actionIdList.clear();
    for(int i=0; i < uiConfig.winactionselector->selectedListWidget()->count(); i++)
    {
        value.setNum(uiConfig.winactionselector->selectedListWidget()->item(i)->type());
        actionIdList.append(value);
    }
    
    // Get the others settings
    if(uiConfig.checkBoxVerticalLayout->isChecked()) {
        isVerticalPref=true;
        m_layout->setOrientation(Qt::Vertical);
    }
    else {
        isVerticalPref=false;
        m_layout->setOrientation(Qt::Horizontal);
    }
    
    cg.writeEntry("idActionsSelected", actionIdList);        
    cg.writeEntry("verticalLayout", isVerticalPref);
    emit configNeedsSaving();

    removeAllActionButtons();
    addAllItems();
    
    updateConstraints();
    
    // resize the plasmoid in planar/mediacenter mode
    if(formFactor() == Plasma::Planar || formFactor() == Plasma::MediaCenter) { 
        if(!isVerticalPref)
          resize((iconList.size()*PLANAR_WIDTH)+(separatorsAdded*SEPARATOR_WIDTH),PLANAR_HEIGHT);
        else
          resize(PLANAR_HEIGHT, (iconList.size()*PLANAR_WIDTH)+(separatorsAdded*SEPARATOR_WIDTH));
    }

}

void ActiveWindowControl::addActionButton(int index, int idAction)
{
        Plasma::IconWidget * icon = new Plasma::IconWidget(KIcon(), "", this);

        switch (idAction)
        {      
                case ACTION_KEEPBELOW :
                icon->setIcon("awc-keepbelow");
                connect(icon, SIGNAL(clicked()), this, SLOT(slotKeepBelow())); break;

                case ACTION_KEEPABOVE :
                icon->setIcon("awc-keepabove");
                connect(icon, SIGNAL(clicked()), this, SLOT(slotKeepAbove())); break;

                case ACTION_SHADE :
                icon->setIcon("awc-shade");
                connect(icon, SIGNAL(clicked()), this, SLOT(slotShade())); break;

                case ACTION_MINIMIZE :
                icon->setIcon("awc-minimize");
                connect(icon, SIGNAL(clicked()), this, SLOT(slotMinimize())); break;

                case ACTION_MAXIMIZE :
                icon->setIcon("awc-maximize");
                connect(icon, SIGNAL(clicked()), this, SLOT(slotMaximize())); break;

                case ACTION_CLOSE :
                icon->setIcon("awc-close");
                connect(icon, SIGNAL(clicked()), this, SLOT(slotClose())); break;
        }

    //icon->hide();  // FIXME  icons looks sometimes small or big one instant when they are created... 
    //could we avoid that ?
    iconList.insert(index, icon);  
    m_layout->addItem(icon);
}

void ActiveWindowControl::addAllItems()
{
    int idAction;
    int adjacentSeparators=0;
    int insertIndex;
    
    separatorsAdded=0;

    // Add all the buttons and separators
    
    for(int i=0; i<actionIdList.size(); i++) {
        idAction=actionIdList[i].toInt();
        
        if(idAction==SEPARATOR)
            adjacentSeparators++;
        
        else {
            if(adjacentSeparators>0) {              
                insertIndex=(i-1)-adjacentSeparators-separatorsAdded;
                if(insertIndex>=0 && insertIndex< m_layout->count())
                    m_layout->setItemSpacing(insertIndex, adjacentSeparators*SEPARATOR_WIDTH);
              
                separatorsAdded+=adjacentSeparators;
                adjacentSeparators=0;
            }
            addActionButton(i, idAction);
        }
    }
    
    // Add the separators at the extrem left or right (if any) by 
    // adjusting the contentMargins
    // NOTE for the Planar and MediaCenter formfactors, we don't use separators at 
    // the beginning or at the end of the buttons list but this is not  usefull anyway... 
  
    int adjacentSepsLeftOrTop=0;
    int adjacentSepsRightOrBottom=0;
    qreal leftmargin, topmargin, rightmargin, bottommargin;
    m_layout->getContentsMargins ( &leftmargin, &topmargin, &rightmargin, &bottommargin );
                                              
    for(int i=0; i<actionIdList.size(); i++) {                          
        if(actionIdList[i].toInt()==SEPARATOR) adjacentSepsLeftOrTop++; else break; }                                                 
    for(int i=actionIdList.size()-1; i>0; i--) {                          
        if(actionIdList[i].toInt()==SEPARATOR) adjacentSepsRightOrBottom++; else break; }     
        
        
    if(formFactor()==Plasma::Horizontal)
        m_layout->setContentsMargins(adjacentSepsLeftOrTop*SEPARATOR_WIDTH, topmargin, 
                                     adjacentSepsRightOrBottom*SEPARATOR_WIDTH, bottommargin);
    else if(formFactor()==Plasma::Vertical)
        m_layout->setContentsMargins(leftmargin, adjacentSepsLeftOrTop*SEPARATOR_WIDTH, 
                                     rightmargin, adjacentSepsRightOrBottom*SEPARATOR_WIDTH);                             
                                    
    separatorsAdded+= adjacentSepsRightOrBottom;                                    
}

void ActiveWindowControl::removeAllActionButtons()  // TODO are all items correctly deleted here ?
{
    Plasma::IconWidget *icon;
    while((icon = dynamic_cast<Plasma::IconWidget*>(m_layout->itemAt(0))) != 0) {
        m_layout->removeAt(0);
        icon->hide();
    }

 //   if(layout()->count() > 1)       // ???
 //       layout()->removeAt(1);
  
    //TODO is takeItem better for this ?
    for(int i=0; i<iconList.size(); i++)  
    {
        Plasma::IconWidget *icon = iconList[i];
        iconList.removeAt(i);
        delete icon;
    }
    iconList.clear();
    
}

//---------------------------------------------------------------------------------------
void ActiveWindowControl::slotClose()
{
    getActiveWindow();
    KWindowInfo info = KWindowSystem::windowInfo(m_activeWindow, NET::WMWindowType);
    NET::WindowType type = info.windowType(NET::NormalMask | NET::DialogMask | NET::OverrideMask |
                          NET::UtilityMask | NET::DesktopMask | NET::DockMask | NET::TopMenuMask | 
                          NET::SplashMask | NET::ToolbarMask | NET::MenuMask);

    if (type != NET::Desktop && type != NET::Dock && type != NET::TopMenu &&
        type != NET::Splash && type != NET::Menu && type != NET::Toolbar ) 
    {
        NETRootInfo i( QX11Info::display(), 0 );
        i.closeWindowRequest(m_activeWindow);
    }

    //TODO: activate next window in the stack
}

void ActiveWindowControl::slotMinimize()
{
    getActiveWindow();
    KWindowSystem::minimizeWindow(m_activeWindow);

    //TODO: activate next window in the stack
}

void ActiveWindowControl::slotMaximize()
{
    getActiveWindow();
    KWindowInfo info = KWindowSystem::windowInfo(m_activeWindow, NET::WMState );
    if(!info.hasState(NET::Max))  KWindowSystem::setState   (m_activeWindow, NET::Max);
    else                          KWindowSystem::clearState (m_activeWindow, NET::Max);

    if(!isOnPanel) KWindowSystem::activateWindow(m_activeWindow);
}   

void ActiveWindowControl::slotShade()
{
    getActiveWindow();
    KWindowInfo info = KWindowSystem::windowInfo(m_activeWindow, NET::WMState );
    if(!info.hasState(NET::Shaded))  KWindowSystem::setState   (m_activeWindow, NET::Shaded);
    else                             KWindowSystem::clearState (m_activeWindow, NET::Shaded);
    
    if(!isOnPanel) KWindowSystem::activateWindow(m_activeWindow);
}

void ActiveWindowControl::slotKeepAbove()
{
    getActiveWindow();
    KWindowInfo info = KWindowSystem::windowInfo(m_activeWindow, NET::WMState );
    if(!info.hasState(NET::KeepAbove))  KWindowSystem::setState   (m_activeWindow, NET::KeepAbove);
    else                                KWindowSystem::clearState (m_activeWindow, NET::KeepAbove);

    if(!isOnPanel) KWindowSystem::activateWindow(m_activeWindow);
}

void ActiveWindowControl::slotKeepBelow()
{
    getActiveWindow();
    KWindowInfo info = KWindowSystem::windowInfo(m_activeWindow, NET::WMState );
    if(!info.hasState(NET::KeepBelow))  KWindowSystem::setState   (m_activeWindow, NET::KeepBelow);
    else                                KWindowSystem::clearState (m_activeWindow, NET::KeepBelow);

    if(!isOnPanel) KWindowSystem::activateWindow(m_activeWindow);
}

void ActiveWindowControl::slotGetWindowTitle() // not used atm
{
//     QList<WId> windows = KWindowSystem::stackingOrder();
//     foreach(WId window, windows) {
//         kDebug()<< "\n--------------------------- " <<
//         KWindowSystem::readNameProperty(window, XA_WM_NAME );
//     }

//     kDebug()<< "\n------------------- " <<KWindowSystem::readNameProperty( KWindowSystem::activeWindow(), XA_WM_NAME );
}

void ActiveWindowControl::getActiveWindow()
{
    if(isOnPanel)
        m_activeWindow=KWindowSystem::activeWindow();
    else
    {
        QList<WId> windows = KWindowSystem::stackingOrder();
        m_activeWindow=windows[windows.size()-3];  //NOTE -3 ?! Don't know why but it seems to work...
        //TODO test when there is only one open window !
    }
}

//void ActiveWindowControl::slotNewActiveWindow(WId id)
//{
    //NOTE this slot is executed two times when the active window change : is this bug ?
    //kDebug()<< "--------------------------- new active win : "<< KWindowSystem::readNameProperty( id, XA_WM_NAME ); 
//}

#include "activewindowcontrol.moc"
