/***************************************************************************
 *   Copyright (C) 2008 by Thomas Schildknecht <danakil@gmail.com>         *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA .        *
 *                                                                         *
 ***************************************************************************/

#ifndef ACTIVEWINDOWCONTROL_H
#define ACTIVEWINDOWCONTROL_H


#include <QGraphicsLinearLayout>

#include <KDialog>

#include <plasma/widgets/iconwidget.h>
#include <Plasma/Applet>
#include <Plasma/Containment>

#include "ui_activewindowcontrolConfig.h"


/**
 * ActiveWindowControl applet
 */
class ActiveWindowControl : public Plasma::Applet
{
    Q_OBJECT

    public:
        /**
         * Constructor
         */
        ActiveWindowControl(QObject *parent, const QVariantList &args);
        /**
         * Destructor
         */
        ~ActiveWindowControl();
        /**
         * Initialization method
         */
        void init();

    public slots:
        /**
         * Overloaded method to show configuration interface
         */
        void showConfigurationInterface();

        void slotActionAdded(QListWidgetItem * item);
        void slotActionRemoved(QListWidgetItem *item);

        void slotClose();
        void slotMaximize();
        void slotMinimize();
        void slotShade();
        void slotKeepAbove();
        void slotKeepBelow();
        void slotGetWindowTitle();


    protected slots:    

        /**
         * Called when the user has clicked OK or apply in the Config interface
         */
        void configAccepted();
        void constraintsEvent(Plasma::Constraints constraints);
        //void slotNewActiveWindow(WId id);

    private:

        void addActionButton(int index, int idAction);
        void addAllItems();
        void removeAllActionButtons();
        void getActiveWindow();


        KDialog *m_configurationDialog;
        Ui::activewindowcontrolConfig uiConfig;
        QList<Plasma::IconWidget*> iconList;

        QStringList actionIdList;
        int separatorsAdded;
        QGraphicsLinearLayout *m_layout;

        bool isOnPanel;
        bool isVerticalPref;
        WId m_activeWindow;        
};

// http://www.ph.unimelb.edu.au/~ssk/kde/srcdoc/kdecore/NETWinInfo.html
// http://api.kde.org/4.x-api/kdelibs-apidocs/kdeui/html/classKWindowSystem.html
// http://api.kde.org/4.x-api/kdelibs-apidocs/kdeui/html/classNET.html#08dce7f5ea8a2a6d1d38aea3498f00ee11465696144d095ebe6d51085f7f47e6

K_EXPORT_PLASMA_APPLET(activewindowcontrol, ActiveWindowControl)

#endif
